@@ -8,12 +8,13 @@ import logging
 import re
 import time
 import gc
+import traceback
 from StringIO import StringIO
 from test import test_support
 
 from concurrent import futures
 from concurrent.futures._base import (
-    PENDING, RUNNING, CANCELLED, CANCELLED_AND_NOTIFIED, FINISHED, Future)
+    PENDING, RUNNING, CANCELLED, CANCELLED_AND_NOTIFIED, FINISHED, Future, BrokenExecutor)
 from concurrent.futures.thread import cpu_count
 
 try:
@@ -28,7 +29,7 @@ def reap_threads(func):
     If threading is unavailable this function does nothing.
     """
     @functools.wraps(func)
-    def decorator(*args): 
+    def decorator(*args):
         key = test_support.threading_setup()
         try:
             return func(*args)
@@ -49,7 +50,7 @@ def _assert_python(expected_success, *args, **env_vars):
     # caller is responsible to pass the full environment.
     if env_vars.pop('__cleanenv', None):
         env = {}
-    env.update(env_vars) 
+    env.update(env_vars)
     cmd_line.extend(args)
     p = subprocess.Popen(cmd_line, stdin=subprocess.PIPE,
                          stdout=subprocess.PIPE, stderr=subprocess.PIPE,
@@ -77,7 +78,7 @@ def assert_python_ok(*args, **env_vars):
     return _assert_python(True, *args, **env_vars)
 
 
-def strip_python_stderr(stderr): 
+def strip_python_stderr(stderr):
     """Strip the stderr of a Python process from potential debug output
     emitted by the interpreter.
 
@@ -205,10 +206,18 @@ class ThreadPoolShutdownTest(ThreadPoolMixin, ExecutorShutdownTest):
         self.executor.submit(mul, 21, 2)
         self.executor.submit(mul, 6, 7)
         self.executor.submit(mul, 3, 14)
+        def acquire_lock(lock):
+            lock.acquire()
+
+        sem = threading.Semaphore(0)
+        for i in range(3):
+            self.executor.submit(acquire_lock, sem)
         self.assertEqual(len(self.executor._threads), 3)
+        for i in range(3):
+            sem.release()
         self.executor.shutdown()
         for t in self.executor._threads:
-            t.join()
+            t.join(5)
 
     def test_context_manager_shutdown(self):
         with futures.ThreadPoolExecutor(max_workers=5) as e:
@@ -217,7 +226,7 @@ class ThreadPoolShutdownTest(ThreadPoolMixin, ExecutorShutdownTest):
                              [5, 4, 3, 2, 1, 0, 1, 2, 3, 4])
 
         for t in executor._threads:
-            t.join()
+            t.join(5)
 
     def test_del_shutdown(self):
         executor = futures.ThreadPoolExecutor(max_workers=5)
@@ -225,9 +234,37 @@ class ThreadPoolShutdownTest(ThreadPoolMixin, ExecutorShutdownTest):
         threads = executor._threads
         del executor
         gc.collect()
+        gc.collect()
 
         for t in threads:
-            t.join()
+            t.join(5)
+
+    def test_thread_names_assigned(self):
+        executor = futures.ThreadPoolExecutor(
+            max_workers=5, thread_name_prefix='SpecialPool')
+        executor.map(abs, range(-5, 5))
+        threads = executor._threads
+        del executor
+        gc.collect()
+        gc.collect()
+
+        for t in threads:
+            self.assertRegexpMatches(t.name, r'^SpecialPool_[0-4]$')
+            t.join(5)
+
+    def test_thread_names_default(self):
+        executor = futures.ThreadPoolExecutor(max_workers=5)
+        executor.map(abs, range(-5, 5))
+        threads = executor._threads
+        del executor
+        gc.collect()
+        gc.collect()
+
+        for t in threads:
+            # Ensure that our default name is reasonably sane and unique when
+            # no thread_name_prefix was supplied.
+            self.assertRegexpMatches(t.name, r'ThreadPoolExecutor-\d+_[0-4]$')
+            t.join(5)
 
 
 class ProcessPoolShutdownTest(ProcessPoolMixin, ExecutorShutdownTest):
@@ -243,7 +280,7 @@ class ProcessPoolShutdownTest(ProcessPoolMixin, ExecutorShutdownTest):
         self.executor.shutdown()
 
         for p in processes:
-            p.join()
+            p.join(5)
 
     def test_context_manager_shutdown(self):
         with futures.ProcessPoolExecutor(max_workers=5) as e:
@@ -252,7 +289,7 @@ class ProcessPoolShutdownTest(ProcessPoolMixin, ExecutorShutdownTest):
                              [5, 4, 3, 2, 1, 0, 1, 2, 3, 4])
 
         for p in processes:
-            p.join()
+            p.join(5)
 
     def test_del_shutdown(self):
         executor = futures.ProcessPoolExecutor(max_workers=5)
@@ -261,10 +298,11 @@ class ProcessPoolShutdownTest(ProcessPoolMixin, ExecutorShutdownTest):
         processes = executor._processes
         del executor
         gc.collect()
+        gc.collect()
 
-        queue_management_thread.join()
+        queue_management_thread.join(5)
         for p in processes:
-            p.join()
+            p.join(5)
 
 
 class WaitTests(unittest.TestCase):
@@ -503,6 +541,49 @@ class ThreadPoolExecutorTest(ThreadPoolMixin, ExecutorTest):
         self.assertEqual(executor._max_workers,
                          (cpu_count() or 1) * 5)
 
+    def test_thread_initializer(self):
+        initialized = []
+        def initializer(i):
+            initialized.append(i)
+        executor = self.executor_type(initializer=initializer, initargs=(1,))
+        executor.submit(time.sleep, 1)
+        executor.submit(time.sleep, 1)
+        executor.shutdown(wait=True)
+        self.assertEqual(initialized, [1, 1])
+
+    def test_broken_thread_initializer(self):
+        def broken_initializer(i):
+            raise ValueError()
+        executor = self.executor_type(initializer=broken_initializer)
+        with self.assertRaises(BrokenExecutor):
+            executor.submit(time.sleep, 1).result()
+        with self.assertRaises(BrokenExecutor):
+            executor.submit(time.sleep, 1)
+        executor.shutdown(wait=True)
+
+    def test_saturation(self):
+        executor = self.executor_type(4)
+        def acquire_lock(lock):
+            lock.acquire()
+
+        sem = threading.Semaphore(0)
+        for i in range(15 * executor._max_workers):
+            executor.submit(acquire_lock, sem)
+        self.assertEqual(len(executor._threads), executor._max_workers)
+        for i in range(15 * executor._max_workers):
+            sem.release()
+        executor.shutdown(wait=True)
+
+    def test_idle_thread_reuse(self):
+        executor = self.executor_type()
+        executor.submit(mul, 21, 2).result()
+        time.sleep(0.001)
+        executor.submit(mul, 6, 7).result()
+        time.sleep(0.001)
+        executor.submit(mul, 3, 14).result()
+        self.assertEqual(len(executor._threads), 1)
+        executor.shutdown(wait=True)
+
 
 class ProcessPoolExecutorTest(ProcessPoolMixin, ExecutorTest):
     pass
@@ -542,22 +623,33 @@ class FutureTests(unittest.TestCase):
     def test_done_callback_raises(self):
         with captured_stderr() as stderr:
             raising_was_called = [False]
+            raising_old_style_was_called = [False]
             fn_was_called = [False]
 
             def raising_fn(callback_future):
                 raising_was_called[0] = True
                 raise Exception('doh!')
 
+            def raising_old_style_fn(callback_future):
+                raising_old_style_was_called[0] = True
+                class OldStyle:  # Does not inherit from object
+                   def __str__(self):
+                       return 'doh!'
+                raise OldStyle()
+
             def fn(callback_future):
                 fn_was_called[0] = True
 
             f = Future()
             f.add_done_callback(raising_fn)
+            f.add_done_callback(raising_old_style_fn)
             f.add_done_callback(fn)
             f.set_result(5)
             self.assertTrue(raising_was_called)
+            self.assertTrue(raising_old_style_was_called)
             self.assertTrue(fn_was_called)
             self.assertIn('Exception: doh!', stderr.getvalue())
+            self.assertIn('OldStyle: doh!', stderr.getvalue())
 
     def test_done_callback_already_successful(self):
         callback_result = [None]
@@ -668,7 +760,7 @@ class FutureTests(unittest.TestCase):
         self.assertEqual(SUCCESSFUL_FUTURE.result(timeout=0), 42)
 
     def test_result_with_success(self):
-        # TODO(brian@sweetapp.com): This test is timing dependant.
+        # TODO(brian@sweetapp.com): This test is timing dependent.
         def notification():
             # Wait until the main thread is waiting for the result.
             time.sleep(1)
@@ -681,7 +773,7 @@ class FutureTests(unittest.TestCase):
         self.assertEqual(f1.result(timeout=5), 42)
 
     def test_result_with_cancel(self):
-        # TODO(brian@sweetapp.com): This test is timing dependant.
+        # TODO(brian@sweetapp.com): This test is timing dependent.
         def notification():
             # Wait until the main thread is waiting for the result.
             time.sleep(1)
@@ -721,6 +813,33 @@ class FutureTests(unittest.TestCase):
 
         self.assertTrue(isinstance(f1.exception(timeout=5), IOError))
 
+    def test_old_style_exception(self):
+        class OldStyle:  # Does not inherit from object
+            def  __str__(self):
+                return 'doh!'
+        callback_exc_info = [None]
+        def fn(callback_future):
+            callback_exc_info[0] = callback_future.exception_info()
+        f = Future()
+        f.add_done_callback(fn)
+        try:
+            raise OldStyle()
+        except OldStyle:
+            want_exc_info = sys.exc_info()
+            f.set_exception_info(*want_exc_info[1:])
+        self.assertEqual(f.exception_info(), want_exc_info[1:])
+        self.assertEqual(callback_exc_info[0], want_exc_info[1:])
+        try:
+            f.result()
+        except OldStyle:
+            got_exc_info = sys.exc_info()
+        else:
+            self.fail('OldStyle exception not raised')
+        self.assertEqual(got_exc_info[:2], want_exc_info[:2])
+        got_tb = traceback.extract_tb(got_exc_info[2])
+        want_tb = traceback.extract_tb(want_exc_info[2])
+        self.assertEqual(got_tb[-len(want_tb):], want_tb)
+
 @reap_threads
 def test_main():
     try:
